from __future__ import annotations

import grp
import io
import os
import pwd
import subprocess
import sys
import time
from collections.abc import Callable
from contextlib import redirect_stderr, redirect_stdout
from typing import Any


try:
    import unshare
except ImportError:
    unshare = None

from ..exceptions import ExecutionFailedError, TimeoutExpiredError
from ..utils.print import print_info


class Sandbox:
    ENV_WHITELIST = ['PATH']

    def __init__(
            self,
            *,
            dry_run: bool = False,
    ) -> None:
        self.dry_run = dry_run

    def _execute_external(
            self,
            command: str | list[str],
            *,
            capture_output: bool = False,
            verbose: bool = False,
            **kwargs: Any,
    ) -> str | None:
        if verbose or self.dry_run:
            if isinstance(command, str):
                cmdline = command
            else:
                cmdline = ' '.join(command)
            if 'preexec_fn' in kwargs:
                cmdline = 'sandbox ' + cmdline
            if 'cwd' in kwargs:
                cmdline = f'cd {kwargs["cwd"]} && {cmdline}'
            print_info('$', cmdline, color='grey')
            print_info('  execution kwargs: ', kwargs, color='grey')

        if self.dry_run:
            return None

        kwargs['check'] = kwargs.get('check', True)  # set check if missing
        try:
            if capture_output:
                start_time = time.monotonic()
                completed_process = subprocess.run(
                    command,
                    close_fds=False,
                    encoding='utf-8',
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,  # https://docs.python.org/3/library/subprocess.html -> capture_output
                    **kwargs
                )
                elapsed_time_seconds = time.monotonic() - start_time
                timeout_msg = ''
                if verbose and 'timeout' in kwargs:
                    timeout_msg = f'\nElapsed time is {elapsed_time_seconds:.2f} ' \
                                  f'with a limit of {kwargs["timeout"]:.0f} seconds\n'
                if completed_process.stdout:
                    return completed_process.stdout + timeout_msg
                return None
            else:
                start_time = time.monotonic()
                subprocess.run(
                    command,
                    close_fds=False,
                    **kwargs
                )
                elapsed_time_seconds = time.monotonic() - start_time
                if verbose and 'timeout' in kwargs:
                    print_info(f'Elapsed time is {elapsed_time_seconds:.2f} '
                               f'with a limit of {kwargs["timeout"]:.0f} seconds')
                return None
        except (subprocess.CalledProcessError, subprocess.TimeoutExpired) as e:
            timeout_msg = ''
            if isinstance(e, subprocess.TimeoutExpired):
                timeout_msg = f'Your solution exceeded time limit: {kwargs["timeout"]} seconds'
                if not capture_output:
                    print_info(timeout_msg, color='red')

            output = e.output or ''
            output = output if isinstance(output, str) else output.decode('utf-8')
            output = output + timeout_msg if capture_output else None
            if isinstance(e, subprocess.TimeoutExpired):
                raise TimeoutExpiredError(output=output) from e
            else:
                raise ExecutionFailedError(output=output) from e

    def _execute_callable(
            self,
            command: Callable[..., Any],
            *,
            capture_output: bool = False,
            verbose: bool = False,
            **kwargs: Any,
    ) -> str | None:
        if verbose or self.dry_run:
            args = ', '.join(f'{k}={repr(v)}' for k, v in sorted(kwargs.items()))
            print_info(f'> {command.__name__}({args})', color='grey')

        if self.dry_run:
            return None

        if capture_output:
            f = io.StringIO()
            with redirect_stdout(f), redirect_stderr(sys.stdout):
                command(**kwargs)
            return f.getvalue()
        else:
            command(**kwargs)
            return None

    def __call__(
            self,
            command: str | list[str] | Callable[..., Any],
            *,
            timeout: float | None = None,
            sandbox: bool = False,
            env_sandbox: bool = False,
            capture_output: bool = False,
            verbose: bool = False,
            **kwargs: Any,
    ) -> str | None:
        if isinstance(command, list) or isinstance(command, str):

            def set_up_env_sandbox() -> None:  # pragma: nocover
                env = os.environ.copy()
                os.environ.clear()
                for variable in self.ENV_WHITELIST:
                    os.environ[variable] = env[variable]

            def set_up_sandbox() -> None:  # pragma: nocover
                set_up_env_sandbox()

                # if unshare:
                #     try:
                #         unshare.unshare(unshare.CLONE_NEWNET)
                #         subprocess.run(['ip', 'link', 'set', 'lo', 'up'], check=True)
                #     except Exception as e:
                #         print_info('WARNING: unable to create new net namespace, running with current one')
                #         if verbose:
                #             print_info(e.__class__.__name__, e)
                # else:
                #     print_info('WARNING: unshare is not installed, running without ip namespace')

                try:
                    uid = pwd.getpwnam('nobody').pw_uid
                    gid = grp.getgrnam('nogroup').gr_gid
                    os.setgroups([])
                    if sys.platform.startswith('linux'):
                        os.setresgid(gid, gid, gid)
                        os.setresuid(uid, uid, uid)
                except Exception as e:
                    print_info('WARNING: UID and GID change failed, running with current user')
                    if verbose:
                        print_info(e.__class__.__name__, e)

                set_up_env_sandbox()

            if env_sandbox:
                kwargs['preexec_fn'] = set_up_env_sandbox
            if sandbox:
                kwargs['preexec_fn'] = set_up_sandbox
            if timeout is not None:
                kwargs['timeout'] = timeout
            return self._execute_external(command, capture_output=capture_output, verbose=verbose, **kwargs)
        elif callable(command):
            if env_sandbox or sandbox or timeout:
                print_info('WARNING: env_sandbox, sandbox and timeout unavailable for callable execution, skip it')
            return self._execute_callable(command, capture_output=capture_output, verbose=verbose, **kwargs)



from __future__ import annotations

import shutil
from pathlib import Path

import git
import pytest

from checker.configs import CheckerTestingConfig
from checker.configs.manytask import ManytaskConfig
from checker.course import Course, FileSystemGroup, FileSystemTask
from checker.exceptions import BadConfig, CheckerException

from .conftest import T_GENERATE_FILE_STRUCTURE


TEST_TIMEZONE = "Europe/Berlin"
TEST_FILE_STRUCTURE = {
    "group1": {
        "task1_1": {".task.yml": "version: 1", "file1_1_1": "", "file1_1_2": "", "extra_file3": ""},
        "task1_2": {".task.yml": "", "file1_2_1": "", "file1_2_2": ""},
        "random_folder": {"file1": "", "file2": ""},
        "extra_file2": "",
        ".group.yml": "",
    },
    "group2": {
        "task2_1": {".task.yml": "", "file2_1_1": "", "file2_1_2": ""},
        "task2_2": {".task.yml": "version: 1"},
        "task2_3": {".task.yml": " \n  \n", "file2_3_1": "", "file2_3_2": "", "file2_3_3": "", "file2_3_4": ""},
        "random_folder": {"file1": "", "file2": ""},
        ".group.yml": "version: 1",
    },
    "group3": {".group.yml": ""},
    "group4": {
        "task4_1": {".task.yml": "version: 1"},
        ".group.yml": "",
    },
    "random_folder": {"file1": "", "file2": ""},
    "root_task_1": {".task.yml": "version: 1", "file1": "", "file2": ""},
    "extra_file1": "",
}
TEST_MANYTASK_CONFIG = ManytaskConfig(
    version=1,
    settings={
        "course_name": "test",
        "gitlab_base_url": "https://google.com",
        "public_repo": "public",
        "students_group": "students",
    },
    ui={"task_url_template": "https://example.com/$GROUP_NAME/$TASK_NAME"},
    deadlines={
        "timezone": TEST_TIMEZONE,
        "schedule": [
            {
                "group": "group1",
                "start": "2020-10-10 00:00:00",
                "end": "200d",
                "enabled": True,
                "tasks": [
                    {"task": "task1_1", "score": 10},
                    {"task": "task1_2", "score": 20},
                ],
            },
            {
                "group": "group2",
                "start": "2020-10-10 00:00:00",
                "end": "200d",
                "enabled": False,
                "tasks": [
                    {"task": "task2_1", "score": 30},
                    {"task": "task2_2", "score": 40},
                    {"task": "task2_3", "score": 50},
                ],
            },
            {
                "group": "group3",
                "start": "2020-10-10 00:00:00",
                "end": "200d",
                "enabled": True,
                "tasks": [],
            },
            {
                "group": "group4",
                "start": "2020-10-10 00:00:00",
                "end": "200d",
                "enabled": True,
                "tasks": [{"task": "task4_1", "score": 50}],
            },
            {
                "group": "group_without_folder",
                "start": "2020-10-10 00:00:00",
                "end": "200d",
                "enabled": True,
                "tasks": [{"task": "root_task_1", "score": 50}],
            },
        ],
    },
)


@pytest.fixture()
def repository_root(generate_file_structure: T_GENERATE_FILE_STRUCTURE) -> Path:
    return generate_file_structure(TEST_FILE_STRUCTURE)


@pytest.fixture()
def git_init_repository_root(repository_root: Path) -> Path:
    # init git repo
    repo = git.Repo.init(repository_root)
    # setup local config
    repo.config_writer().set_value("user", "name", "test_user").release()
    repo.config_writer().set_value("user", "email", "not@val.id").release()
    # commit changes
    repo.git.add(".")
    repo.git.commit("-m", "initial commit")
    return repository_root


class TestCourse:
    def test_init(self, repository_root: Path) -> None:
        test_course = Course(manytask_config=TEST_MANYTASK_CONFIG, repository_root=repository_root)
        assert test_course.repository_root == repository_root
        assert test_course.manytask_config == TEST_MANYTASK_CONFIG

    def test_validate(self, repository_root: Path) -> None:
        test_course = Course(manytask_config=TEST_MANYTASK_CONFIG, repository_root=repository_root)

        try:
            test_course.validate()
        except Exception as e:
            pytest.fail(f"Validation failed: {e}")

    def test_search_for_groups_by_configs(self, repository_root: Path) -> None:
        potential_groups = list(Course._search_for_groups_by_configs(repository_root))
        assert len(potential_groups) == 4
        assert sum(len(group.tasks) for group in potential_groups) == 6
        for group in potential_groups:
            assert isinstance(group, FileSystemGroup)
            assert (repository_root / group.relative_path).exists()

    def test_search_for_tasks_by_configs(self, repository_root: Path) -> None:
        tasks = list(Course._search_for_tasks_by_configs(repository_root))
        assert len(tasks) == 7
        for task in tasks:
            assert isinstance(task, FileSystemTask)
            assert (repository_root / task.relative_path).exists()

    def test_validate_missing_task(self, repository_root: Path) -> None:
        shutil.rmtree(repository_root / "group1" / "task1_1")
        with pytest.raises(BadConfig):
            Course(manytask_config=TEST_MANYTASK_CONFIG, repository_root=repository_root).validate()

    def test_validate_missing_group(self, repository_root: Path) -> None:
        shutil.rmtree(repository_root / "group3")
        with pytest.warns():
            Course(manytask_config=TEST_MANYTASK_CONFIG, repository_root=repository_root).validate()

    def test_init_task_bad_config(self, repository_root: Path) -> None:
        with open(repository_root / "group1" / "task1_1" / Course.TASK_CONFIG_NAME, "w") as f:
            f.write("bad_config")

        with pytest.raises(BadConfig):
            Course(manytask_config=TEST_MANYTASK_CONFIG, repository_root=repository_root)

    @pytest.mark.parametrize(
        "enabled, expected_num_groups",
        [
            (None, 4),
            (True, 3),
            (False, 1),
        ],
    )
    def test_get_groups(self, enabled: bool | None, expected_num_groups, repository_root: Path) -> None:
        test_course = Course(manytask_config=TEST_MANYTASK_CONFIG, repository_root=repository_root)

        groups = test_course.get_groups(enabled=enabled)
        assert isinstance(groups, list)
        assert all(isinstance(group, FileSystemGroup) for group in groups)
        assert len(groups) == expected_num_groups

    @pytest.mark.parametrize(
        "enabled, expected_num_tasks",
        [
            (None, 7),
            (True, 4),
            (False, 3),
        ],
    )
    def test_get_tasks(self, enabled: bool | None, expected_num_tasks, repository_root: Path) -> None:
        test_course = Course(manytask_config=TEST_MANYTASK_CONFIG, repository_root=repository_root)

        tasks = test_course.get_tasks(enabled=enabled)
        assert isinstance(tasks, list)
        assert all(isinstance(task, FileSystemTask) for task in tasks)
        assert len(tasks) == expected_num_tasks

    def test_detect_changes_not_a_repo(self, repository_root: Path) -> None:
        test_course = Course(manytask_config=TEST_MANYTASK_CONFIG, repository_root=repository_root)
        with pytest.raises(CheckerException):
            test_course.detect_changes(CheckerTestingConfig.ChangesDetectionType.COMMIT_MESSAGE)

    @pytest.mark.parametrize(
        "branch_name, changed_files, expected_changed_tasks",
        [
            ("task1_1", ["group1/task1_1/file1_1_1"], ["task1_1"]),
            ("task1_1", ["group1/task1_1/file1_1_1", "random_file.txt", "group1/task1_1/file1_1_1"], ["task1_1"]),
            ("task2_1", ["group2/task2_1/file1_1_1"], []),  # not enabled
            ("not_a_task", ["group2/task2_1/file2_1_1"], []),
            ("root_task_1", ["root_task_1/file1"], ["root_task_1"]),
            ("root_task_1", [], ["root_task_1"]),
            # group names
            ("group1", ["group1/task1_1/file1_1_1"], ["task1_1", "task1_2"]),
            ("group1", ["group2/task2_1/file1_1_1"], ["task1_1", "task1_2"]),
            ("group1", [], ["task1_1", "task1_2"]),
            ("group2", ["group2/task2_1/file1_1_1"], []),  # not enabled
            ("group3", ["group1/task1_1/file1_1_1"], []),  # empty group
            ("group_without_folder", [], ["root_task_1"]),
        ],
    )
    def test_detect_changes_by_branch_name(
        self,
        git_init_repository_root: Path,
        branch_name: str,
        changed_files: list[str],
        expected_changed_tasks: list[str],
    ) -> None:
        test_course = Course(manytask_config=TEST_MANYTASK_CONFIG, repository_root=git_init_repository_root)
        repo = git.Repo(git_init_repository_root)

        # create new branch
        repo.git.checkout("-b", branch_name)
        # create or change files
        for filename in changed_files:
            Path(git_init_repository_root / filename).write_text(f"random_text_to_write_in_file {filename}")
        # commit changes (allow empty commit)
        repo.git.add(".")
        repo.git.commit("-m", "random commit message", "--allow-empty")

        changed_tasks = test_course.detect_changes(CheckerTestingConfig.ChangesDetectionType.BRANCH_NAME)
        assert isinstance(changed_tasks, list)
        assert all(isinstance(task, FileSystemTask) for task in changed_tasks)
        assert len(changed_tasks) == len(expected_changed_tasks)
        assert sorted(task.name for task in changed_tasks) == sorted(expected_changed_tasks)

    @pytest.mark.parametrize(
        "commit_message, changed_files, expected_changed_tasks",
        [
            ("task1_1", ["group1/task1_1/file1_1_1"], ["task1_1"]),
            ("fixses in task1_1", ["group1/task1_1/file1_1_1"], ["task1_1"]),
            (
                "task1_1 some commit",
                ["group1/task1_1/file1_1_1", "random_file.txt", "group1/task1_1/file1_1_1"],
                ["task1_1"],
            ),
            ("add fixes for task2_1", ["group2/task2_1/file1_1_1"], []),  # not enabled
            ("    not_a_task", ["group2/task2_1/file2_1_1"], []),
            ("root_task_1", ["root_task_1/file1"], ["root_task_1"]),
            ("my root_task_1 is really cool", [], ["root_task_1"]),
            (
                "my root_task_1 and task1_1 and not enabled task2_1",
                ["group2/task2_1/file2_1_1"],
                ["root_task_1", "task1_1"],
            ),
            ("commit root_task_1", [], ["root_task_1"]),
            ("commit root_task_1 and some more", [], ["root_task_1"]),
            # group names
            ("commit with group1 group1 group1 group1", ["group1/task1_1/file1_1_1"], ["task1_1", "task1_2"]),
            ("commit with group1", ["group2/task2_1/file1_1_1"], ["task1_1", "task1_2"]),
            ("my solutions for group1 and group_without_folder", [], ["task1_1", "task1_2", "root_task_1"]),
            ("group2 and group22", ["group2/task2_1/file1_1_1"], []),  # not enabled
            ("group3", ["group1/task1_1/file1_1_1"], []),  # empty group
            ("group_without_folder", [], ["root_task_1"]),
        ],
    )
    def test_detect_changes_by_commit_message(
        self,
        git_init_repository_root: Path,
        commit_message: str,
        changed_files: list[str],
        expected_changed_tasks: list[str],
    ) -> None:
        test_course = Course(manytask_config=TEST_MANYTASK_CONFIG, repository_root=git_init_repository_root)
        repo = git.Repo(git_init_repository_root)

        # create or change files
        for filename in changed_files:
            Path(git_init_repository_root / filename).write_text(f"random_text_to_write_in_file {filename}")
        # commit changes (allow empty commit)
        repo.git.add(".")
        repo.git.commit("-m", commit_message, "--allow-empty")

        changed_tasks = test_course.detect_changes(CheckerTestingConfig.ChangesDetectionType.COMMIT_MESSAGE)
        assert isinstance(changed_tasks, list)
        assert all(isinstance(task, FileSystemTask) for task in changed_tasks)
        assert len(changed_tasks) == len(expected_changed_tasks)
        assert sorted(task.name for task in changed_tasks) == sorted(expected_changed_tasks)

    @pytest.mark.parametrize(
        "changed_files, expected_changed_tasks",
        [
            (["group1/task1_1/file.txt"], ["task1_1"]),
            (["group1/task1_1/file.txt", "random_file.txt", "group1/task1_1/file.txt"], ["task1_1"]),
            (["group2/task2_1/file.txt"], []),  # not enabled
            (["group2/task2_1/file2_1_1.txt"], []),  # not enabled
            (["some_root_file", "random_folder/random_file.txt"], []),
            (
                ["group2/task2_1/file2_1_1.txt", "group1/task1_1/file.txt", "root_task_1/some.txt"],
                ["task1_1", "root_task_1"],
            ),
            (["root_task_1/file1.txt"], ["root_task_1"]),
            ([], []),
        ],
    )
    def test_detect_changes_by_last_commit_changes(
        self, git_init_repository_root: Path, changed_files: list[str], expected_changed_tasks: list[str]
    ) -> None:
        test_course = Course(manytask_config=TEST_MANYTASK_CONFIG, repository_root=git_init_repository_root)
        repo = git.Repo(git_init_repository_root)

        # create or change files
        for filename in changed_files:
            Path(git_init_repository_root / filename).write_text(f"random_text_to_write_in_file {filename}")
        # commit changes (allow empty commit)
        repo.git.add(".")
        repo.git.commit("-m", "random commit message", "--allow-empty")

        changed_tasks = test_course.detect_changes(CheckerTestingConfig.ChangesDetectionType.LAST_COMMIT_CHANGES)
        assert isinstance(changed_tasks, list)
        assert all(isinstance(task, FileSystemTask) for task in changed_tasks)
        assert len(changed_tasks) == len(expected_changed_tasks)
        assert sorted(task.name for task in changed_tasks) == sorted(expected_changed_tasks)



from __future__ import annotations

import sys
from typing import Any


def print_info(
        *args: Any,
        file: Any = None,
        color: str | None = None,
        **kwargs: Any,
) -> None:
    colors = {
        'white': '\033[97m',
        'cyan': '\033[96m',
        'pink': '\033[95m',
        'blue': '\033[94m',
        'orange': '\033[93m',
        'green': '\033[92m',
        'red': '\033[91m',
        'grey': '\033[90m',
        'endc': '\033[0m',
    }

    file = file or sys.stderr

    data = ' '.join(map(str, args))
    if color in colors:
        print(colors[color] + data + colors['endc'], file=file, **kwargs)
    else:
        print(data, file=file, **kwargs)
    file.flush()


def print_header_info(
        info_string: str,
) -> None:
    info_extended_string = ' ' + info_string + ' '
    info_extended_string = '+' * (50 - len(info_extended_string) // 2) + \
                           info_extended_string + \
                           '+' * (50 - len(info_extended_string) // 2)
    print_info('')
    print_info('+' * len(info_extended_string), color='pink')
    print_info(info_extended_string, color='pink')
    print_info('+' * len(info_extended_string), color='pink')


def print_task_info(
        task_name: str,
) -> None:
    print_header_info(f'Testing tasks: <{task_name}>')



"""Helpers to interact with manytask (push scores tasks)"""
from __future__ import annotations

import io
import json
import time
from datetime import datetime

import requests

from ..exceptions import GetFailedError, PushFailedError


def push_report(
        report_base_url: str,
        tester_token: str,
        task_name: str,
        user_id: int,
        score: float,
        files: dict[str, tuple[str, io.BufferedReader]] | None = None,
        send_time: datetime | None = None,
        check_deadline: bool = True,
        use_demand_multiplier: bool = True,
) -> tuple[str, int, str | None, str | None, float | None]:
    # Do not expose token in logs.
    data = {
        'token': tester_token,
        'task': task_name,
        'user_id': user_id,
        'score': score,
        'check_deadline': check_deadline,
        'use_demand_multiplier': use_demand_multiplier,
    }
    if send_time:
        data['commit_time'] = send_time

    response = None
    for _ in range(3):
        response = requests.post(url=f'{report_base_url}/api/report', data=data, files=files)

        if response.status_code < 500:
            break
        time.sleep(1.0)
    assert response is not None

    if response.status_code >= 500:
        response.raise_for_status()
        assert False, 'Not Reachable'  # pragma: no cover
    elif response.status_code >= 400:
        # Client error often means early submission
        raise PushFailedError(f'{response.status_code}: {response.text}')
    else:
        try:
            result = response.json()
            result_commit_time = result.get('commit_time', None)
            result_submit_time = result.get('submit_time', None)
            demand_multiplier = float(result.get('demand_multiplier', 1))
            return result['username'], result['score'], result_commit_time, result_submit_time, demand_multiplier
        except (json.JSONDecodeError, KeyError) as e:
            raise PushFailedError('Unable to decode response') from e


def get_score(
        report_base_url: str,
        tester_token: str,
        task_name: str,
        user_id: int,
) -> float | None:
    # Do not expose token in logs.
    data = {
        'token': tester_token,
        'task': task_name,
        'user_id': user_id,
    }
    # response = None
    for _ in range(3):
        response = requests.get(url=f'{report_base_url}/api/score', data=data)

        if response.status_code < 500:
            break
        time.sleep(1.0)

    if response.status_code >= 500:
        response.raise_for_status()
        assert False, 'Not Reachable'  # pragma: no cover
    # Client error often means early submission
    elif response.status_code >= 400:
        raise GetFailedError(f'{response.status_code}: {response.text}')
    else:
        try:
            result = response.json()
            return result['score']
        except (json.JSONDecodeError, KeyError):
            # raise GetFailedError()
            pass

    return None



from __future__ import annotations

from dataclasses import dataclass


# Base exception for all package
class CheckerException(Exception):
    __test__ = False  # to disable pytest detecting it as Test class
    pass


# All deadlines and conf exceptions
class BadConfig(CheckerException):
    pass


class BadTaskConfig(BadConfig):
    pass


class BadGroupConfig(BadConfig):
    pass


# Tests exceptions
class TesterException(CheckerException):
    pass


class TesterNotImplemented(TesterException):
    pass


class TaskTesterException(TesterException):
    pass


class TaskTesterTestConfigException(TaskTesterException):
    pass


# Tests exceptions (with output)
@dataclass
class RunFailedError(TesterException):
    msg: str = ''
    output: str | None = None

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}: {self.msg}'


class ExecutionFailedError(RunFailedError):
    pass


class TimeoutExpiredError(ExecutionFailedError):
    pass


class BuildFailedError(RunFailedError):
    pass


class RegexpCheckFailedError(RunFailedError):
    pass


class StylecheckFailedError(RunFailedError):
    pass


class TestsFailedError(RunFailedError):
    pass


# Manytask exceptions
class ManytaskRequestFailedError(CheckerException):
    pass


class PushFailedError(ManytaskRequestFailedError):
    pass


class GetFailedError(ManytaskRequestFailedError):
    pass



from .checker import (  # noqa: F401
    CheckerConfig,
    CheckerExportConfig,
    CheckerStructureConfig,
    CheckerSubConfig,
    CheckerTestingConfig,
    PipelineStageConfig,
)
from .manytask import ManytaskConfig, ManytaskDeadlinesConfig, ManytaskGroupConfig, ManytaskTaskConfig  # noqa: F401



"""
All course configurations
Include tests, layout, manytask url, gitlab urls, etc. settings
"""
from __future__ import annotations

import os
from dataclasses import InitVar, dataclass
from pathlib import Path

import yaml

from ..exceptions import BadConfig
from ..utils.print import print_info


@dataclass
class CourseConfig:
    # main course settings
    name: str

    # course
    deadlines: str

    # checker
    system: str
    templates: str  # create or explicit

    # manytask
    manytask_url: str

    # gitlab
    course_group: str
    public_repo: str
    students_group: str
    lectures_repo: str | None = None
    default_branch: str = 'main'
    gitlab_url: str = 'https://gitlab.manytask.org'
    gitlab_service_username: str = 'manytask'
    gitlab_service_email: str = 'no-reply@gitlab.manytask.org'
    gitlab_service_name: str = 'Manytask Bot'

    # course default
    second_deadline_max: float = 0.5
    low_demand_bonus_bound: float = 1.0
    max_low_demand_bonus: float = 1.0

    # checker default
    layout: str = 'groups'
    executor: str = 'sandbox'
    tester_path: str | None = None

    # info
    links: dict[str, str] | None = None

    # credentials
    manytask_token: str | None = None
    gitlab_service_token: str | None = None
    gitlab_api_token: str | None = None

    manytask_token_id: InitVar[str] = 'TESTER_TOKEN'
    gitlab_service_token_id: InitVar[str] = 'GITLAB_SERVICE_TOKEN'
    gitlab_api_token_id: InitVar[str] = 'GITLAB_API_TOKEN'

    def __post_init__(
            self,
            manytask_token_id: str,
            gitlab_service_token_id: str,
            gitlab_api_token_id: str,
    ) -> None:
        self.manytask_token = os.environ.get(manytask_token_id)
        if not self.manytask_token:
            print_info(f'Unable to find env <{manytask_token_id}>', color='orange')

        self.gitlab_service_token = os.environ.get(gitlab_service_token_id)
        if not self.gitlab_service_token:
            print_info(f'Unable to find env <{gitlab_service_token_id}>', color='orange')

        self.gitlab_api_token = os.environ.get(gitlab_api_token_id)
        if not self.gitlab_api_token:
            print_info(f'Unable to find env <{gitlab_api_token_id}>', color='orange')

    @classmethod
    def from_yaml(cls, course_config: Path) -> 'CourseConfig':
        try:
            with open(course_config) as config_file:
                config_dict = yaml.safe_load(config_file)
        except (yaml.YAMLError, FileNotFoundError) as e:
            raise BadConfig(f'Unable to load deadlines config file <{course_config}>') from e

        try:
            return cls(**config_dict)
        except (KeyError, TypeError, ValueError) as e:
            raise BadConfig('Invalid course config') from e



from __future__ import annotations

import re
import shutil
import subprocess
from pathlib import Path

from .print import print_info


def filename_match_patterns(
        file: Path,
        patterns: list[str],
) -> bool:
    """
    Check if filename match any of patterns given
    @param file: Path object to check
    @param patterns: list of regexp pattern (?)
    @return: True if any of patterns applicable
    """
    for pattern in patterns:
        if file.match(pattern):
            return True
    return False


def copy_files(
        source: Path | None,
        target: Path,
        patterns: list[str] | None = None,
        ignore_patterns: list[str] | None = None,
) -> None:
    """
    Copy files between 2 directories
    @param source: Directory or file to copy from (none to skip)
    @param target: Directory or file to copy to
    @param patterns: Patterns to copy
    @param ignore_patterns: Patterns to ignore during copy
    @return: None
    """
    ignore_patterns = ignore_patterns or []
    target.mkdir(parents=True, exist_ok=True)

    if source is None:
        print_info(f'Warning: Skip copying files from <{source}> to <{target}>')
        return

    ignore_files: list[Path] = sum([
        list(source.glob(ignore_pattern))
        for ignore_pattern in ignore_patterns
    ], [])
    for pattern in (patterns or ['*']):
        for file in source.glob(pattern):
            if file in ignore_files:
                continue
            relative_filename = str(file.relative_to(source))
            source_path = source / relative_filename
            target_path = target / relative_filename
            if file.is_dir():
                copy_files(
                    source_path, target_path,
                    patterns=['*'],
                    ignore_patterns=ignore_patterns,
                )
                continue

            shutil.copyfile(str(source_path), str(target_path))


def check_file_contains_regexp(
        filename: Path,
        regexps: list[str],
) -> bool:
    """
    Check regexps appears in the file
    @param filename: Filename to check
    @param regexps: list of forbidden regexp
    @raise AssertionError: if file does not exist
    @return: True if any of regexp found
    """
    assert filename.exists() and filename.is_file()

    with open(filename, 'r', encoding='utf-8') as f:
        file_content = f.read()

        for regexp in regexps:
            if re.search(regexp, file_content, re.MULTILINE):
                return True

    return False


def check_folder_contains_regexp(
        folder: Path,
        extensions: list[str],
        regexps: list[str],
        raise_on_found: bool = False,
) -> bool:
    """
    Check regexps appears in any file in the folder
    @param folder: Folder to check
    @param extensions: Source files extensions
    @param regexps: list of forbidden regexp
    @param raise_on_found: Raise Error on Found Exception
    @raise AssertionError: if folder or any file does not exist
    @return: True if any of regexp found
    """
    assert folder.exists() and folder.is_dir()

    for source_path in folder.glob('**/*.*'):
        if any(str(source_path).endswith(ext) for ext in extensions):
            if check_file_contains_regexp(source_path, regexps):
                if raise_on_found:
                    raise AssertionError(f'File <{source_path}> contains one of <{regexps}>')
                return True
    return False


def check_files_contains_regexp(
        folder: Path,
        regexps: list[str],
        patterns: list[str] | None = None,
        raise_on_found: bool = False,
) -> bool:
    """
    Check regexps appears in files that matching patterns
    @param folder: Folder to check
    @param regexps: list of forbidden regexp
    @param patterns: list of patterns for file matching
    @param raise_on_found: Raise Error on Found Exception
    @raise AssertionError: if folder or any file does not exist
    @return: True if any of regexp found
    """
    assert folder.exists() and folder.is_dir()
    if patterns is None:
        patterns = ['**/*.*']
    for pattern in patterns:
        for source_path in folder.glob(pattern):
            if check_file_contains_regexp(source_path, regexps):
                if raise_on_found:
                    raise AssertionError(f'File <{source_path}> contains one of <{regexps}>')
                return True
    return False


def get_folders_diff(
        old_folder: Path,
        new_folder: Path,
        skip_binary: bool = True,
        exclude_patterns: list[str] | None = None,
) -> list[str]:
    """
    Return diff files between 2 folders
    @param old_folder: Old folder
    @param new_folder: New folder with some changes files, based on old folder
    @param skip_binary: Skip binary files
    @param exclude_patterns: Exclude files that match pattern
    @return: list of changed files as strings
    """
    # diff docs https://www.gnu.org/software/diffutils/manual/html_node/diff-Options.html
    # -N/--new-file - If one file is missing, treat it as present but empty
    # -w/--ignore-all-space - ignore all spaces and tabs  e.g. if ( a == b)  is equal to if(a==b)
    # -r/--recursive - recursively compare any subdirectories found
    # -q/--brief - report only when files differ
    # --strip-trailing-cr - strip trailing carriage return on input
    # -x/--exclude [pattern] - exclude files that match pattern

    # TODO: check format options to work, or --left-column options
    exclude_args = [f'--exclude={pattern}' for pattern in exclude_patterns] if exclude_patterns else []
    # exclude_args = []
    result = subprocess.run(
        [
            'diff',
            '--brief',
            '--recursive',
            '--new-file',
            '--strip-trailing-cr',
            *exclude_args,
            old_folder.absolute(),
            new_folder.absolute()
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    output = result.stdout.decode()

    # TODO: make it work with whitespace in filenames

    changed = []
    for line in output.split('\n'):
        if line.startswith('Only in'):
            assert False, 'Will be treated as change due to --new-file option'
        elif line.startswith('Files'):
            _, file1, _, file2, _ = line.split()
            changed.append(Path(file2).relative_to(new_folder))
        elif line.startswith('Binary files'):
            if skip_binary:
                continue
            _, _, file1, _, file2, _ = line.split()
            changed.append(Path(file2).relative_to(new_folder))

    return [str(i) for i in changed]


def get_folders_diff_except_public(
        public_folder: Path,
        old_folder: Path,
        new_folder: Path,
        skip_binary: bool = True,
        exclude_patterns: list[str] | None = None,
) -> list[str]:
    """
    Return diff files between 2 folders except files that are equal to public folder files
    @param public_folder: Public folder
    @param old_folder: Old folder
    @param new_folder: New folder with some changes files, based on old folder
    @param skip_binary: Skip binary files
    @param exclude_patterns: Exclude files that match pattern
    @return: list of changed files as strings
    """

    changed_files_old_new = get_folders_diff(
        old_folder,
        new_folder,
        skip_binary=skip_binary,
        exclude_patterns=exclude_patterns,
    )
    changed_files_public_new = get_folders_diff(
        public_folder,
        new_folder,
        skip_binary=skip_binary,
        exclude_patterns=exclude_patterns,
    )

    # TODO: Remove logging
    print_info('\nchanged_files_old_new:', color='grey')
    for i in changed_files_old_new:
        print_info(f'  {i}', color='grey')
    print_info('\nchanged_files_public_new:', color='grey')
    for i in changed_files_public_new:
        print_info(f'  {i}', color='grey')

    return [
        str(i)
        for i in set(changed_files_old_new) & set(changed_files_public_new)
    ]



from __future__ import annotations

from typing import Optional, Union

from ..exceptions import PluginExecutionFailed
from .base import PluginABC, PluginOutput


class RunScriptPlugin(PluginABC):
    """Plugin for running scripts."""

    name = "run_script"

    class Args(PluginABC.Args):
        origin: str
        script: Union[str, list[str]]  # as pydantic does not support | in older python versions
        timeout: Union[float, None] = None  # as pydantic does not support | in older python versions
        env_whitelist: Optional[list[str]] = None

    def _run(self, args: Args, *, verbose: bool = False) -> PluginOutput:  # type: ignore[override]
        import subprocess

        def set_up_env_sandbox() -> None:  # pragma: nocover
            import os

            if args.env_whitelist is None:
                return

            env = os.environ.copy()
            os.environ.clear()
            for variable in args.env_whitelist:
                os.environ[variable] = env[variable]

        if isinstance(args.script, list):
            safe_shell_script = " ".join(args.script)
        else:
            safe_shell_script = args.script

        try:
            result = subprocess.run(
                safe_shell_script,
                shell=True,
                cwd=args.origin,
                timeout=args.timeout,  # kill process after timeout, raise TimeoutExpired
                check=True,  # raise CalledProcessError if return code is non-zero
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,  # merge stderr & stdout to single output
                preexec_fn=set_up_env_sandbox if args.env_whitelist else None,
            )
        except (subprocess.CalledProcessError, subprocess.TimeoutExpired) as e:
            output = e.output or ""
            output = output if isinstance(output, str) else output.decode("utf-8")

            if isinstance(e, subprocess.TimeoutExpired):
                raise PluginExecutionFailed(
                    f"Script timed out after {e.timeout}s ({args.timeout}s limit)",
                    output=output,
                ) from e
            else:
                raise PluginExecutionFailed(
                    f"Script failed with exit code {e.returncode}",
                    output=output,
                ) from e

        return PluginOutput(
            output=result.stdout.decode("utf-8"),
        )
